// -*- C++ -*-
/*!
 * @file  BirdsEyeComposer.cpp
 * @brief 8枚の画像から1枚の俯瞰画像へ合成を行う
 * @date $Date$
 *
 * @author 株式会社 東日本計算センター
 *
 * $Id$
 */

#include "BirdsEyeComposer.h"
#include <coil/Async.h>

// Module specification
// <rtc-template block="module_spec">
static const char* birdseyecomposer_spec[] =
  {
    "implementation_id", "BirdsEyeComposer",
    "type_name",         "BirdsEyeComposer",
    "description",       "8枚の画像から1枚の俯瞰画像へ合成を行う",
    "version",           "1.0",
    "vendor",            "EAC",
    "category",          "カメラ",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    // Configuration variables
    "conf.default.Angle_B", "-24",
    "conf.default.Angle_D", "-158",
    "conf.default.Angle_E", "-180",
    "conf.default.Angle_F", "-202",
    "conf.default.Angle_H", "-340",
    "conf.default.Coord_X_A", "135",
    "conf.default.Coord_Y_A", "0",
    "conf.default.Coord_X_B", "210",
    "conf.default.Coord_Y_B", "23",
    "conf.default.Coord_X_C", "337",
    "conf.default.Coord_Y_C", "160",
    "conf.default.Coord_X_D", "202",
    "conf.default.Coord_Y_D", "382",
    "conf.default.Coord_X_E", "135",
    "conf.default.Coord_Y_E", "400",
    "conf.default.Coord_X_F", "30",
    "conf.default.Coord_Y_F", "382",
    "conf.default.Coord_X_G", "0",
    "conf.default.Coord_Y_G", "160",
    "conf.default.Coord_X_H", "35",
    "conf.default.Coord_Y_H", "20",
    "conf.default.Display_A", "On",
    "conf.default.Display_B", "On",
    "conf.default.Display_C", "On",
    "conf.default.Display_D", "On",
    "conf.default.Display_E", "On",
    "conf.default.Display_F", "On",
    "conf.default.Display_G", "On",
    "conf.default.Display_H", "On",
    // Widget
    "conf.__widget__.Angle_B", "slider.1",
    "conf.__widget__.Angle_D", "slider.1",
    "conf.__widget__.Angle_E", "slider.1",
    "conf.__widget__.Angle_F", "slider.1",
    "conf.__widget__.Angle_H", "slider.1",
    "conf.__widget__.Coord_X_A", "slider.1",
    "conf.__widget__.Coord_Y_A", "slider.1",
    "conf.__widget__.Coord_X_B", "slider.1",
    "conf.__widget__.Coord_Y_B", "slider.1",
    "conf.__widget__.Coord_X_C", "slider.1",
    "conf.__widget__.Coord_Y_C", "slider.1",
    "conf.__widget__.Coord_X_D", "slider.1",
    "conf.__widget__.Coord_Y_D", "slider.1",
    "conf.__widget__.Coord_X_E", "slider.1",
    "conf.__widget__.Coord_Y_E", "slider.1",
    "conf.__widget__.Coord_X_F", "slider.1",
    "conf.__widget__.Coord_Y_F", "slider.1",
    "conf.__widget__.Coord_X_G", "slider.1",
    "conf.__widget__.Coord_Y_G", "slider.1",
    "conf.__widget__.Coord_X_H", "slider.1",
    "conf.__widget__.Coord_Y_H", "slider.1",
    "conf.__widget__.Display_A", "radio",
    "conf.__widget__.Display_B", "radio",
    "conf.__widget__.Display_C", "radio",
    "conf.__widget__.Display_D", "radio",
    "conf.__widget__.Display_E", "radio",
    "conf.__widget__.Display_F", "radio",
    "conf.__widget__.Display_G", "radio",
    "conf.__widget__.Display_H", "radio",
    // Constraints
    "conf.__constraints__.Angle_B", "-360<=x<=360",
    "conf.__constraints__.Angle_D", "-360<=x<=360",
    "conf.__constraints__.Angle_E", "-360<=x<=360",
    "conf.__constraints__.Angle_F", "-360<=x<=360",
    "conf.__constraints__.Angle_H", "-360<=x<=360",
    "conf.__constraints__.Coord_X_A", "0<=x<=257",
    "conf.__constraints__.Coord_Y_A", "0<=x<=400",
    "conf.__constraints__.Coord_X_B", "0<=x<=257",
    "conf.__constraints__.Coord_Y_B", "0<=x<=400",
    "conf.__constraints__.Coord_X_C", "0<=x<=337",
    "conf.__constraints__.Coord_Y_C", "0<=x<=320",
    "conf.__constraints__.Coord_X_D", "0<=x<=257",
    "conf.__constraints__.Coord_Y_D", "0<=x<=400",
    "conf.__constraints__.Coord_X_E", "0<=x<=257",
    "conf.__constraints__.Coord_Y_E", "0<=x<=400",
    "conf.__constraints__.Coord_X_F", "0<=x<=257",
    "conf.__constraints__.Coord_Y_F", "0<=x<=400",
    "conf.__constraints__.Coord_X_G", "0<=x<=337",
    "conf.__constraints__.Coord_Y_G", "0<=x<=320",
    "conf.__constraints__.Coord_X_H", "0<=x<=257",
    "conf.__constraints__.Coord_Y_H", "0<=x<=400",
    "conf.__constraints__.Display_A", "(On,Off)",
    "conf.__constraints__.Display_B", "(On,Off)",
    "conf.__constraints__.Display_C", "(On,Off)",
    "conf.__constraints__.Display_D", "(On,Off)",
    "conf.__constraints__.Display_E", "(On,Off)",
    "conf.__constraints__.Display_F", "(On,Off)",
    "conf.__constraints__.Display_G", "(On,Off)",
    "conf.__constraints__.Display_H", "(On,Off)",
    ""
  };
// </rtc-template>

/*!
 * @brief constructor
 * @param manager Maneger Object
 */
BirdsEyeComposer::BirdsEyeComposer(RTC::Manager* manager)
	// <rtc-template block="initializer">
	: RTC::DataFlowComponentBase(manager),
	m_MultiCaptureimageIn("MultiCaptureDataIn", m_MultiCaptureimage),
	m_BirdsEyeimageOut("BirdsEyeDataOut", m_BirdsEyeimage),
	m_BirdsEyeControlServPort("BirdsEyeControlServ"),
	m_BirdsEyeSystemContServPort("BirdsEyeSystemContServ"),

	//add
	m_uWBackUpComposeMode(COMPOSE_MODE_OFF),
	m_uWBackUpCaptureMode(CAPT_MODE_VIDEO),
	m_uWBackUpCaptureTrigger(CAPT_TRIG_STOP),
	m_uWFileBranchNum(0),
	m_uWResetCnt(1),
	m_uWReqReset(RESET_TRIG_OFF),
	m_blErrSts(FALSE)

	// </rtc-template>
{
	::memset(&m_stBackUpFileCreateTime, 0, sizeof(tm));
}

/*!
 * @brief destructor
 */
BirdsEyeComposer::~BirdsEyeComposer()
{
}


/*!
 * 俯瞰画像合成RTCの初期化処理
 */
RTC::ReturnCode_t BirdsEyeComposer::onInitialize()
{
#ifdef _DEBUG
	std::cout << "[debug]BirdsEyeCompose::onInitialize()" << std::endl;
#endif

	// Registration: InPort/OutPort/Service
	// <rtc-template block="registration">
	// Set InPort buffers
	addInPort("MultiCaptureDataIn", m_MultiCaptureimageIn);

	// Set OutPort buffer
	addOutPort("BirdsEyeDataOut", m_BirdsEyeimageOut);

	// Set service provider to Ports
	m_BirdsEyeControlServPort.registerProvider("BirdsEyeComposeMode", "BirdsEyeControlServ::BirdsEyeComposeMode", m_BirdsEyeComposeModeProv);
	m_BirdsEyeControlServPort.registerProvider("BirdsEyeCaptureMode", "BirdsEyeControlServ::BirdsEyeCaptureMode", m_BirdsEyeCaptureModeProv);
	m_BirdsEyeControlServPort.registerProvider("BirdsEyeCaptureTrigger", "BirdsEyeControlServ::BirdsEyeCaptureTrigger", m_BirdsEyeCaptureTriggerProv);
	m_BirdsEyeControlServPort.registerProvider("BirdsEyeProgress", "BirdsEyeControlServ::BirdsEyeProgress", m_BirdsEyeProgressProv);
	m_BirdsEyeSystemContServPort.registerProvider("BirdsEyeReset", "BirdsEyeSystemContServ::BirdsEyeReset", m_BirdsEyeResetProv);

	// Set service consumers to Ports

	// Set CORBA Service Ports
	addPort(m_BirdsEyeControlServPort);
	addPort(m_BirdsEyeSystemContServPort);

	// </rtc-template>

	// <rtc-template block="bind_config">
	// Bind variables and configuration variable
	bindParameter("Angle_B", m_Angle_B, "-24");
	bindParameter("Angle_D", m_Angle_D, "-158");
	bindParameter("Angle_E", m_Angle_E, "-180");
	bindParameter("Angle_F", m_Angle_F, "-202");
	bindParameter("Angle_H", m_Angle_H, "-340");
	bindParameter("Coord_X_A", m_Coord_X_A, "135");
	bindParameter("Coord_Y_A", m_Coord_Y_A, "0");
	bindParameter("Coord_X_B", m_Coord_X_B, "210");
	bindParameter("Coord_Y_B", m_Coord_Y_B, "23");
	bindParameter("Coord_X_C", m_Coord_X_C, "337");
	bindParameter("Coord_Y_C", m_Coord_Y_C, "160");
	bindParameter("Coord_X_D", m_Coord_X_D, "202");
	bindParameter("Coord_Y_D", m_Coord_Y_D, "382");
	bindParameter("Coord_X_E", m_Coord_X_E, "135");
	bindParameter("Coord_Y_E", m_Coord_Y_E, "400");
	bindParameter("Coord_X_F", m_Coord_X_F, "30");
	bindParameter("Coord_Y_F", m_Coord_Y_F, "382");
	bindParameter("Coord_X_G", m_Coord_X_G, "0");
	bindParameter("Coord_Y_G", m_Coord_Y_G, "160");
	bindParameter("Coord_X_H", m_Coord_X_H, "35");
	bindParameter("Coord_Y_H", m_Coord_Y_H, "20");
	bindParameter("Display_A", m_Display_A, "On");
	bindParameter("Display_B", m_Display_B, "On");
	bindParameter("Display_C", m_Display_C, "On");
	bindParameter("Display_D", m_Display_D, "On");
	bindParameter("Display_E", m_Display_E, "On");
	bindParameter("Display_F", m_Display_F, "On");
	bindParameter("Display_G", m_Display_G, "On");
	bindParameter("Display_H", m_Display_H, "On");
	// </rtc-template>

	return RTC::RTC_OK;
}

/*!
 * 俯瞰画像合成RTCの終了処理
 */

RTC::ReturnCode_t BirdsEyeComposer::onFinalize()
{
#ifdef _DEBUG
	std::cout << "[debug]BirdsEyeCompose::onFinalize()" << std::endl;
#endif

	return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t BirdsEyeComposer::onStartup(RTC::UniqueId ec_id)
{
	return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t BirdsEyeComposer::onShutdown(RTC::UniqueId ec_id)
{
	return RTC::RTC_OK;
}
*/

/*!
 * 俯瞰画像合成RTCの開始処理
 */

RTC::ReturnCode_t BirdsEyeComposer::onActivated(RTC::UniqueId ec_id)
{
#ifdef _DEBUG
	std::cout << "[debug]BirdsEyeCompose::onActivated()" << std::endl;
#endif

	//add
	uWord_Type uWIndex = 0;

	//接続最大数分の入力画像領域を確保
	m_MultiCaptureimage.images.length(NM33_CONNECT_ENABLE_MAX);
	for(uWIndex = 0; uWIndex < NM33_CONNECT_ENABLE_MAX; uWIndex++){
		m_MultiCaptureimage.images[uWIndex].width = 0;
		m_MultiCaptureimage.images[uWIndex].height = 0;
	}

	//出力画像データの初期化
	m_BirdsEyeimage.width = 0;
	m_BirdsEyeimage.height = 0;


	return RTC::RTC_OK;
}

/*!
 * 俯瞰画像合成RTCの停止処理
 */

RTC::ReturnCode_t BirdsEyeComposer::onDeactivated(RTC::UniqueId ec_id)
{
#ifdef _DEBUG
	std::cout << "[debug]BirdsEyeCompose::onDeactivated()" << std::endl;
#endif

	//add
	m_uWBackUpComposeMode = COMPOSE_MODE_OFF;
	m_uWBackUpCaptureMode = CAPT_MODE_VIDEO;
	m_uWBackUpCaptureTrigger = CAPT_TRIG_STOP;
	m_uWFileBranchNum = 0;
	::memset(&m_stBackUpFileCreateTime, 0, sizeof(tm));

	return RTC::RTC_OK;
}

/*!
 * 俯瞰画像合成を行う周期処理
 */

RTC::ReturnCode_t BirdsEyeComposer::onExecute(RTC::UniqueId ec_id)
{
#ifdef _DEBUG
//	std::cout << "[debug]BirdsEyeCompose::onExecute()" << std::endl;
#endif

	uWord_Type uWResult = RESULT_OK;
	uWord_Type uWMatrixMode = CAPTURE_NOACTION;
	uWord_Type uWLoopCnt = 0;
	Bool_Type blBirdsEyeFlag = FALSE;
	uWord_Type uWCapWidth = 0;
	uWord_Type uWCapHeight = 0;

	//マトリクスチェック
	uWResult = BirdsEyeActMatrix(&uWMatrixMode);
	if(uWResult == RESULT_NG){
		return RTC::RTC_ERROR;
	}

	if((uWMatrixMode == CAPTURE_ONCE) || (uWMatrixMode == CAPTURE_CYCLE)){
		if(m_MultiCaptureimageIn.isNew()){
			//データポートに更新がかかればデータポートを読み込む
			m_MultiCaptureimageIn.read();
			blBirdsEyeFlag = TRUE;
		}
	}

	if(blBirdsEyeFlag == TRUE){
		//データポートの入力画像データのチェック
		for(uWLoopCnt = 0; uWLoopCnt < NM33_CONNECT_ENABLE_MAX; uWLoopCnt++){
			uWCapWidth = m_MultiCaptureimage.images[ uWLoopCnt ].width;
			uWCapHeight = m_MultiCaptureimage.images[ uWLoopCnt ].height;

			if((uWCapWidth <= 0) || (uWCapHeight <= 0)){
				//8台分すべての画像データが揃っていない場合エラーメッセージを出力する。（次Phaseで接続台数は可変になる為ここではエラー状態に遷移させない）
				std::cout << MES_BIRDS_EYE_INPUT_DATA_NON << " = [" << uWLoopCnt << "]. " << "[" << uWCapWidth << "]. " << "[" << uWCapHeight << "]" << std::endl;
				return RTC::RTC_OK;
			}
		}
		
		//コンフィギュレーションの範囲チェック
		uWResult = BirdsEyeConfigChk();
		if(uWResult == RESULT_NG){
			return RTC::RTC_ERROR;
		}

		//俯瞰合成処理を実行
		uWResult = MakeBirdsEyeCompose();
		if(uWResult == RESULT_NG){
			return RTC::RTC_ERROR;
		}

		//キャプチャモードの場合、合成画像をJPGファイルに保存する
		if(uWMatrixMode == CAPTURE_ONCE){
			uWResult = BirdsEyeJPGFileCreate();
			if(uWResult == RESULT_NG){
				return RTC::RTC_ERROR;
			}
		}
		//データポートを出力する
		m_BirdsEyeimageOut.write();

		//FPSのログ出力
		BirdsEyeFPSLogOut();
	}

	//ResetによるRetryカウントをクリアする
	m_uWResetCnt = 1;

	return RTC::RTC_OK;

}

/*!
 * 俯瞰画像合成RTCのエラー遷移処理
 */

RTC::ReturnCode_t BirdsEyeComposer::onAborting(RTC::UniqueId ec_id)
{
#ifdef _DEBUG
	std::cout << "[debug]BirdsEyeCompose::onAborting()" << std::endl;
#endif

	return RTC::RTC_OK;
}

/*!
 * 俯瞰画像合成RTCのエラー処理
 */

RTC::ReturnCode_t BirdsEyeComposer::onError(RTC::UniqueId ec_id)
{
	m_uWReqReset = m_BirdsEyeResetProv.GetBirdsEyeReset();

	if((m_uWReqReset == RESET_TRIG_ON) && (m_uWResetCnt <= RESET_RETRY_MAX_CNT)){
		//RTCの状態をResetへ遷移させる
		get_context(ec_id)->reset_component(getObjRef());
		std::cout << MES_RECOVERY_RESET_RETRY_CNT << m_uWResetCnt << std::endl;
	}else{
		//エラー状態が確定している
		m_blErrSts = TRUE;
	}

	return RTC::RTC_OK;
}

/*!
 * 俯瞰画像合成RTCのReset処理
 */

RTC::ReturnCode_t BirdsEyeComposer::onReset(RTC::UniqueId ec_id)
{
#ifdef _DEBUG
	std::cout << "[debug]BirdsEyeCompose::onReset()" << std::endl;
#endif

	//リカバリResetのRetry処理
	if((m_uWReqReset == RESET_TRIG_ON) && (m_uWResetCnt <= RESET_RETRY_MAX_CNT)){
		m_uWResetCnt++;
		// 非同期処理にてRTCの状態をactivateへ遷移させる
		coil::AsyncInvoker(this, activator(ec_id), true)->invoke();
	}

	//RT System Editor状態でReset操作した時に動作する。既にエラー状態が確定している場合カウントをクリアする
	if(m_blErrSts == TRUE){
		//リカバリReset Retry回数を初期化
		m_uWResetCnt = 1;

		//エラー状態を初期化
		m_blErrSts = FALSE;
	}

	return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t BirdsEyeComposer::onStateUpdate(RTC::UniqueId ec_id)
{
	return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t BirdsEyeComposer::onRateChanged(RTC::UniqueId ec_id)
{
	return RTC::RTC_OK;
}
*/

/*******************************************************************************
	サブルーチン
********************************************************************************/
/*******************************************************************************
 *	Function Name		：BirdsEyeActMatrix
 *	Description			：動作マトリクスの確認を行い動作を決定する
 *	Argument			：Out：マトリクスモード
 *	Return				：正常終了/異常終了
 *******************************************************************************/
uWord_Type BirdsEyeComposer::BirdsEyeActMatrix(uWord_Type *uWMatrixMode)
{
	//マトリクステーブル
	uWord_Type BirdsEyeActMaxtixTbl[2][2]={
			//キャプチャ			//ムービー
		{	CAPTURE_NOACTION	,	CAPTURE_STOP	},	//停止
		{	CAPTURE_ONCE		,	CAPTURE_CYCLE	},	//開始
	};

	//サービスポートから最新の変換モード/撮影モード/動作トリガを取得
	uWord_Type uWCurrentCompMode = m_BirdsEyeComposeModeProv.GetBirdsEyeComposeMode();
	uWord_Type uWCurrentCapMode = m_BirdsEyeCaptureModeProv.GetBirdsEyeCaptureMode();
	uWord_Type uWCurrentTrigMode = m_BirdsEyeCaptureTriggerProv.GetBirdsEyeCaptureTrigger();

	Bool_Type blMatrixCheck = FALSE;
	uWord_Type uWReturn = RESULT_OK;

	if(uWMatrixMode == NULL){
		return (RESULT_NG);
	}else{
		*uWMatrixMode = CAPTURE_NOACTION;
	}

	//範囲チェック
	if((uWCurrentCompMode >= COMPOSE_MODE_MAX) || (uWCurrentCapMode >= CAPT_MODE_MAX) || (uWCurrentTrigMode >= CAPT_TRIG_MAX)){
		return (RESULT_NG);
	}

	//変化チェック
	if(m_uWBackUpComposeMode != uWCurrentCompMode){
		m_uWBackUpComposeMode = uWCurrentCompMode;
	}

	if(m_uWBackUpCaptureMode != uWCurrentCapMode){
		m_uWBackUpCaptureMode = uWCurrentCapMode;
	}

	if(m_uWBackUpCaptureTrigger != uWCurrentTrigMode){
		m_uWBackUpCaptureTrigger = uWCurrentTrigMode;
		//動作トリガに変化がある場合、無条件でマトリクスの確認を行う
		blMatrixCheck = TRUE;
	}else{
		if(m_uWBackUpCaptureMode == CAPT_MODE_VIDEO){
			//動画モードはサイクルで動作する為、常にマトリクスの確認を行う
			blMatrixCheck = TRUE;
		}else{
			//静止画モードは、1回のみ動作の為動作トリガが変化した時のみマトリクスの確認を行う
		}
	}

	//動作トリガに変化があればマトリクスの確認を行う
	if((blMatrixCheck == TRUE) && (m_uWBackUpComposeMode == COMPOSE_MODE_ON)){
		*uWMatrixMode = BirdsEyeActMaxtixTbl[uWCurrentTrigMode][uWCurrentCapMode];
	}

	if(*uWMatrixMode == CAPTURE_ONCE){
		//静止画モードは一回のみの動作の為トリガをOFFする
		m_BirdsEyeCaptureTriggerProv.ReqBirdsEyeCaptureTrigger(CAPT_TRIG_STOP);
		m_uWBackUpCaptureTrigger = CAPT_TRIG_STOP;
	}

	return (uWReturn);
}

/*******************************************************************************
 *	Function Name		：MakeBirdsEyeCompose
 *	Description			：俯瞰合成処理（会津大学提供Code）
 *	Argument			：なし
 *	Return				：正常終了/異常終了
 *******************************************************************************/
uWord_Type BirdsEyeComposer::MakeBirdsEyeCompose(void)
{
	//データポートから8枚の画像データを取得
	cv::Mat src_A = cv::Mat(cv::Size(m_MultiCaptureimage.images[0].width, m_MultiCaptureimage.images[0].height), CV_8UC3);
	::memcpy(src_A.data, (void *)&(m_MultiCaptureimage.images[0].pixels[0]), m_MultiCaptureimage.images[0].pixels.length());

	cv::Mat src_B = cv::Mat(cv::Size(m_MultiCaptureimage.images[1].width, m_MultiCaptureimage.images[1].height), CV_8UC3);
	::memcpy(src_B.data, (void *)&(m_MultiCaptureimage.images[1].pixels[0]), m_MultiCaptureimage.images[1].pixels.length());

	cv::Mat src_C = cv::Mat(cv::Size(m_MultiCaptureimage.images[2].width, m_MultiCaptureimage.images[2].height), CV_8UC3);
	::memcpy(src_C.data, (void *)&(m_MultiCaptureimage.images[2].pixels[0]), m_MultiCaptureimage.images[2].pixels.length());

	cv::Mat src_D = cv::Mat(cv::Size(m_MultiCaptureimage.images[3].width, m_MultiCaptureimage.images[3].height), CV_8UC3);
	::memcpy(src_D.data, (void *)&(m_MultiCaptureimage.images[3].pixels[0]), m_MultiCaptureimage.images[3].pixels.length());

	cv::Mat src_E = cv::Mat(cv::Size(m_MultiCaptureimage.images[4].width, m_MultiCaptureimage.images[4].height), CV_8UC3);
	::memcpy(src_E.data, (void *)&(m_MultiCaptureimage.images[4].pixels[0]), m_MultiCaptureimage.images[4].pixels.length());

	cv::Mat src_F = cv::Mat(cv::Size(m_MultiCaptureimage.images[5].width, m_MultiCaptureimage.images[5].height), CV_8UC3);
	::memcpy(src_F.data, (void *)&(m_MultiCaptureimage.images[5].pixels[0]), m_MultiCaptureimage.images[5].pixels.length());

	cv::Mat src_G = cv::Mat(cv::Size(m_MultiCaptureimage.images[6].width, m_MultiCaptureimage.images[6].height), CV_8UC3);
	::memcpy(src_G.data, (void *)&(m_MultiCaptureimage.images[6].pixels[0]), m_MultiCaptureimage.images[6].pixels.length());

	cv::Mat src_H = cv::Mat(cv::Size(m_MultiCaptureimage.images[7].width, m_MultiCaptureimage.images[7].height ), CV_8UC3);
	::memcpy(src_H.data, (void *)&(m_MultiCaptureimage.images[7].pixels[0]), m_MultiCaptureimage.images[7].pixels.length());

	//透視投影変換行列を求める
	cv::Point2f src_point[4], dst_point[4];

	src_point[0] = cv::Point2f(137.0, 172.0);
	src_point[1] = cv::Point2f(182.0, 172.0);
	src_point[2] = cv::Point2f(172.0, 150.0);
	src_point[3] = cv::Point2f(146.0, 150.0);

	dst_point[0] = cv::Point2f(137.0, 172.0);
	dst_point[1] = cv::Point2f(182.0, 172.0);
	dst_point[2] = cv::Point2f(182.0, 127.0);
	dst_point[3] = cv::Point2f(137.0, 127.0);

	cv::Mat map_matrix = cv::getPerspectiveTransform(src_point, dst_point);

//	cout << map_matrix << endl;

	//透視投影変換行列を用いて画像を変換する
	cv::Mat dst_A;
	cv::Mat dst_B;
	cv::Mat dst_C;
	cv::Mat dst_D;
	cv::Mat dst_E;
	cv::Mat dst_F;
	cv::Mat dst_G;
	cv::Mat dst_H;

	warpPerspective(src_A, dst_A, map_matrix, src_A.size(), cv::INTER_LINEAR);
	warpPerspective(src_B, dst_B, map_matrix, src_B.size(), cv::INTER_LINEAR);
	warpPerspective(src_C, dst_C, map_matrix, src_B.size(), cv::INTER_LINEAR);
	warpPerspective(src_D, dst_D, map_matrix, src_B.size(), cv::INTER_LINEAR);
	warpPerspective(src_E, dst_E, map_matrix, src_B.size(), cv::INTER_LINEAR);
	warpPerspective(src_F, dst_F, map_matrix, src_B.size(), cv::INTER_LINEAR);
	warpPerspective(src_G, dst_G, map_matrix, src_B.size(), cv::INTER_LINEAR);
	warpPerspective(src_H, dst_H, map_matrix, src_B.size(), cv::INTER_LINEAR);

	//出力結果を回転
	float angle, scale = 1.0;
	cv::Point2f center;
	cv::Mat R;

	//画像Aはそのまま
	cv::Mat new_dst_A = dst_A;

	//画像Bは-25度回転
//	angle = -24;
	angle = m_Angle_B;
	center = cv::Point2f(dst_B.cols * 0.5, dst_B.rows * 0.5);
	R = cv::getRotationMatrix2D(center, angle, scale);
	cv::Mat new_dst_B;
	warpAffine(dst_B, new_dst_B, R, dst_B.size());

	//画像Cは90度回転		
	cv::Mat new_dst_C(320, 240, CV_8UC3);
	for(int v = 0; v < dst_C.rows; v++){
		for(int u = 0; u < dst_C.cols; u++){
			new_dst_C.at<cv::Vec3b>(u, 239 - v) = dst_C.at<cv::Vec3b>(v, u);
		}
	}

	//画像Dは155度回転
//	angle = -158;
	angle = m_Angle_D;
	center = cv::Point2f(dst_D.cols * 0.5, dst_D.rows * 0.5);
	R = cv::getRotationMatrix2D(center, angle, scale);
	cv::Mat new_dst_D;
	warpAffine(dst_D, new_dst_D, R, dst_D.size());

	//画像Eは180度回転
//	angle = -180;
	angle = m_Angle_E;
	center = cv::Point2f(dst_E.cols * 0.5, dst_E.rows * 0.5);
	R = cv::getRotationMatrix2D(center, angle, scale);
	cv::Mat new_dst_E;
	warpAffine(dst_E, new_dst_E, R, dst_E.size());

	//画像Fは205度回転
//	angle = -202;
	angle = m_Angle_F;
	center = cv::Point2f(dst_F.cols * 0.5, dst_F.rows * 0.5);
	R = cv::getRotationMatrix2D(center, angle, scale);
	cv::Mat new_dst_F;
	warpAffine(dst_F, new_dst_F, R, dst_F.size());

	//画像Gは270度回転
	cv::Mat new_dst_G(320, 240, CV_8UC3);
	for(int v = 0; v < dst_G.rows; v++){
		for(int u = 0; u < dst_G.cols; u++){
			new_dst_G.at<cv::Vec3b>(319 - u, v) = dst_G.at<cv::Vec3b>(v, u);
		}
	}

	//画像Hは340度回転
//	angle = -340;
	angle = m_Angle_H;
	center = cv::Point2f(dst_H.cols * 0.5, dst_H.rows * 0.5);
	R = cv::getRotationMatrix2D(center, angle, scale);
	cv::Mat new_dst_H;
	warpAffine(dst_H, new_dst_H, R, dst_H.size());

	//画像を1枚にまとめる
	cv::Mat Z(640, 577, CV_8UC3);

	//Cを合成(右)
	if(m_Display_C == DISPLAY_ON){
		for(int v = 0; v < new_dst_C.rows; v++){
			for(int u = 0; u < new_dst_C.cols; u++){
				Z.at<cv::Vec3b>((v + m_Coord_Y_C), (u + m_Coord_X_C)) = new_dst_C.at<cv::Vec3b>(v, u);
			}
		}
	}

	//Gを合成(左)
	if(m_Display_G == DISPLAY_ON){
		for(int v = 0; v < new_dst_G.rows; v++){
			for(int u = 0; u < new_dst_G.cols; u++){
				Z.at<cv::Vec3b>((v + m_Coord_Y_G), (u + m_Coord_X_G)) = new_dst_G.at<cv::Vec3b>(v, u);
			}
		}
	}

	//Aを合成(前)
	if(m_Display_A == DISPLAY_ON){
		for(int v = 0; v < new_dst_A.rows; v++){
			for(int u = 0; u < new_dst_A.cols; u++){
				Z.at<cv::Vec3b>((v + m_Coord_Y_A), (u + m_Coord_X_A)) = new_dst_A.at<cv::Vec3b>(v, u);
			}
		}
	}

	//Eを合成(後)
	if(m_Display_E == DISPLAY_ON){
		for(int v = 0; v < new_dst_E.rows; v++){
			for(int u = 0; u < new_dst_E.cols; u++){
				Z.at<cv::Vec3b>((v + m_Coord_Y_E), (u + m_Coord_X_E)) = new_dst_E.at<cv::Vec3b>(v, u);
			}
		}
	}

	//Hを合成(左前)
	if(m_Display_H == DISPLAY_ON){
		for(int v = 0; v < new_dst_H.rows; v++){
			for(int u = 0; u < new_dst_H.cols; u++){
				//灰色又は黒色の時のみ画像をコピーする
				if((Z.at<cv::Vec3b>((v + m_Coord_Y_H), (u + m_Coord_X_H))[0] == 205) &&
					(Z.at<cv::Vec3b>((v + m_Coord_Y_H), (u + m_Coord_X_H))[1] == 205) &&
					(Z.at<cv::Vec3b>((v + m_Coord_Y_H), (u + m_Coord_X_H))[2] == 205))
				{
					Z.at<cv::Vec3b>((v + m_Coord_Y_H), (u + m_Coord_X_H)) = new_dst_H.at<cv::Vec3b>(v, u);
				}else if((Z.at<cv::Vec3b>((v + m_Coord_Y_H), (u + m_Coord_X_H))[0] == 0) &&
						(Z.at<cv::Vec3b>((v + m_Coord_Y_H), (u + m_Coord_X_H))[1] == 0) &&
						(Z.at<cv::Vec3b>((v + m_Coord_Y_H), (u + m_Coord_X_H))[2] == 0))
				{
					Z.at<cv::Vec3b>((v + m_Coord_Y_H), (u + m_Coord_X_H)) = new_dst_H.at<cv::Vec3b>(v, u);
				}
			}
		}
	}

	//Bを合成(右前)
	if(m_Display_B == DISPLAY_ON){
		for(int v = 0; v < new_dst_B.rows; v++){
			for(int u = 0; u < new_dst_B.cols; u++){
				//灰色又は黒色の時のみ画像をコピーする
				if ((Z.at<cv::Vec3b>((v + m_Coord_Y_B), (u + m_Coord_X_B))[0] == 205) &&
					(Z.at<cv::Vec3b>((v + m_Coord_Y_B), (u + m_Coord_X_B))[1] == 205) &&
					(Z.at<cv::Vec3b>((v + m_Coord_Y_B), (u + m_Coord_X_B))[2] == 205))
				{
					Z.at<cv::Vec3b>((v + m_Coord_Y_B), (u + m_Coord_X_B)) = new_dst_B.at<cv::Vec3b>(v, u);
				}else if((Z.at<cv::Vec3b>((v + m_Coord_Y_B), (u + m_Coord_X_B))[0] == 0) && 
						(Z.at<cv::Vec3b>((v + m_Coord_Y_B), (u + m_Coord_X_B))[1] == 0) && 
						(Z.at<cv::Vec3b>((v + m_Coord_Y_B), (u + m_Coord_X_B))[2] == 0))
				{
					Z.at<cv::Vec3b>((v + m_Coord_Y_B), (u + m_Coord_X_B)) = new_dst_B.at<cv::Vec3b>(v, u);
				}
			}
		}
	}

	//Fを合成(左後)
	if(m_Display_F == DISPLAY_ON){
		for(int v = 0; v < new_dst_D.rows; v++){
			for(int u = 0; u < new_dst_F.cols; u++){
				//灰色又は黒色の時のみ画像をコピーする
				if((Z.at<cv::Vec3b>((v + m_Coord_Y_F), (u + m_Coord_X_F))[0] == 205) &&
					(Z.at<cv::Vec3b>((v + m_Coord_Y_F), (u + m_Coord_X_F))[1] == 205) &&
					(Z.at<cv::Vec3b>((v + m_Coord_Y_F), (u + m_Coord_X_F))[2] == 205))
				{
					Z.at<cv::Vec3b>((v + m_Coord_Y_F), (u + m_Coord_X_F)) = new_dst_F.at<cv::Vec3b>(v, u);
				}else if((Z.at<cv::Vec3b>((v + m_Coord_Y_F), (u + m_Coord_X_F))[0] == 0) && 
					(Z.at<cv::Vec3b>((v + m_Coord_Y_F), (u + m_Coord_X_F))[1] == 0) && 
					(Z.at<cv::Vec3b>((v + m_Coord_Y_F), (u + m_Coord_X_F))[2] == 0))
				{
					Z.at<cv::Vec3b>((v + m_Coord_Y_F), (u + m_Coord_X_F)) = new_dst_F.at<cv::Vec3b>(v, u);
				}
			}
		}
	}

	//Dを合成
	if(m_Display_D == DISPLAY_ON){
		for(int v = 0; v < new_dst_D.rows; v++){
			for(int u = 0; u < new_dst_D.cols; u++){
				//灰色又は黒色の時のみ画像をコピーする
				if((Z.at<cv::Vec3b>((v + m_Coord_Y_D), (u + m_Coord_X_D))[0] == 205) &&
					(Z.at<cv::Vec3b>((v + m_Coord_Y_D), (u + m_Coord_X_D))[1] == 205) &&
					(Z.at<cv::Vec3b>((v + m_Coord_Y_D), (u + m_Coord_X_D))[2] == 205))
				{
					Z.at<cv::Vec3b>((v + m_Coord_Y_D), (u + m_Coord_X_D)) = new_dst_D.at<cv::Vec3b>(v, u);
				}else if((Z.at<cv::Vec3b>((v + m_Coord_Y_D), (u + m_Coord_X_D))[0] == 0) && 
					(Z.at<cv::Vec3b>((v + m_Coord_Y_D), (u + m_Coord_X_D))[1] == 0) && 
					(Z.at<cv::Vec3b>((v + m_Coord_Y_D), (u + m_Coord_X_D))[2] == 0))
				{
					Z.at<cv::Vec3b>((v + m_Coord_Y_D), (u + m_Coord_X_D)) = new_dst_D.at<cv::Vec3b>(v, u);
				}
			}
		}
	}

	//黒を消す
	for(int v = 0; v < Z.rows; v++){
		for(int u = 0; u < Z.cols; u++){
			if(Z.at<cv::Vec3b>(v, u)[0] == 0 && Z.at<cv::Vec3b>(v, u)[1] == 0 && Z.at<cv::Vec3b>(v, u)[2] == 0){
				Z.at<cv::Vec3b>(v, u)[0] = 205;
				Z.at<cv::Vec3b>(v, u)[1] = 205;
				Z.at<cv::Vec3b>(v, u)[2] = 205;
			}
		}
	}

	cv::Mat newZ(COMPOSE_FRAME_WIDTH, COMPOSE_FRAME_HEIGHT, CV_8UC3);

	//生成された俯瞰画像の中央部を切り取る
	for(int v = 71; v < 566; v++){
		for(int u = 35; u < 506; u++){
			newZ.at<cv::Vec3b>(v - 71, u - 35) = Z.at<cv::Vec3b>(v, u);
		}
	}

	//俯瞰合成したデータのチェック
	if(( newZ.cols <= 0 ) || ( newZ.rows <= 0 )){
		std::cout << MES_BIRDS_EYE_COMPOSE_ERROR << " = [" << newZ.cols << "]. " << "[" << newZ.rows << "]" << std::endl;
		return RESULT_NG;
	}

	//CameraImage型の出力ポートへデータを設定
	sWord_Type sWlength = ((newZ.channels() * newZ.cols) * newZ.rows);
	m_BirdsEyeimage.width = newZ.cols;
	m_BirdsEyeimage.height = newZ.rows;
	m_BirdsEyeimage.pixels.length(sWlength);
	::memcpy((void *)&(m_BirdsEyeimage.pixels[0]), newZ.data, sWlength);	//データのコピー

	return RESULT_OK;
}

/*******************************************************************************
 *	Function Name		：BirdsEyeJPGFileCreate
 *	Description			：合成画像をJPGファイルへ保存する
 *	Argument			：なし
 *	Return				：正常終了/異常終了
 *******************************************************************************/
uWord_Type BirdsEyeComposer::BirdsEyeJPGFileCreate(void)
{
	//キャプチャ開始前の時間を保存
	time_t stNow = time(NULL);
	struct tm *stNow_p = localtime(&stNow);

	uWord_Type uWRet = ::memcmp(&m_stBackUpFileCreateTime, stNow_p, sizeof(tm));

	if(uWRet == 0){
		//一致していれば枝番をインクリメント
		m_uWFileBranchNum ++;
	}else{
		m_uWFileBranchNum = 0;
	}

	::memcpy(&m_stBackUpFileCreateTime, stNow_p, sizeof(tm));

	//JPG画像格納フォルダの生成
	DWORD dWResult = GetFileAttributes("BirdsEyeOutPut");
	if (dWResult == -1){
		//格納フォルダが無ければを作成
		BOOL bRet = CreateDirectory("BirdsEyeOutPut", NULL);
		if(bRet == FALSE){
			//フォルダの生成に失敗
			std::cout << MES_BIRDS_EYE_FOLDER_CREATE_ERROR << std::endl;
			return RESULT_NG;
		}
	}

	//出力ポートから画像データを取得
	cv::Mat ImageData = cv::Mat(cv::Size(m_BirdsEyeimage.width, m_BirdsEyeimage.height), CV_8UC3);
	::memcpy(ImageData.data, (void *)&(m_BirdsEyeimage.pixels[0]), m_BirdsEyeimage.pixels.length());

	//作成するJPGファイルの名称を付ける
	Char_Type cFilename[128] = "";

	sprintf(cFilename, "BirdsEyeOutPut/BirdsEye_%04d%02d%02d%02d%02d%02d_%d.jpg", (m_stBackUpFileCreateTime.tm_year + 1900), (m_stBackUpFileCreateTime.tm_mon + 1), 
			 m_stBackUpFileCreateTime.tm_mday, m_stBackUpFileCreateTime.tm_hour, m_stBackUpFileCreateTime.tm_min, m_stBackUpFileCreateTime.tm_sec, m_uWFileBranchNum);

	//JPGファイル生成
	BOOL bRet2 = imwrite(cFilename, ImageData);
	if(bRet2 == FALSE){
		//ファイル書き込みに失敗
		std::cout << MES_BIRDS_EYE_JPG_FILE_CREATE_ERROR << std::endl;
		return RESULT_NG;
	}

	return RESULT_OK;
}

/*******************************************************************************
 *	Function Name		：BirdsEyeConfigChk
 *	Description			：コンフィギュレーション設定値のチェック処理
 *	Argument			：なし
 *	Return				：正常終了/異常終了
 *******************************************************************************/
uWord_Type BirdsEyeComposer::BirdsEyeConfigChk(void)
{
	struct ConfigData_t{
		sWord_Type sWConfig;		//コンフィギュレーション設定
		sWord_Type sWConfigDataMin;	//Min値
		sWord_Type sWConfigDataMax;	//Max値
	};

	//角度、配置座標調整値チェックテーブル
	struct ConfigData_t ConfigDataChkTbl[] = {
		m_Angle_B	,	ANGLE_MIN		,	ANGLE_MAX			,
		m_Angle_D	,	ANGLE_MIN		,	ANGLE_MAX			,
		m_Angle_E	,	ANGLE_MIN		,	ANGLE_MAX			,
		m_Angle_F	,	ANGLE_MIN		,	ANGLE_MAX			,
		m_Angle_H	,	ANGLE_MIN		,	ANGLE_MAX			,
		m_Coord_X_A	,	HORIZONTAL_X_MIN,	HORIZONTAL_X_MAX	,
		m_Coord_Y_A	,	HORIZONTAL_Y_MIN,	HORIZONTAL_Y_MAX	,
		m_Coord_X_B	,	HORIZONTAL_X_MIN,	HORIZONTAL_X_MAX	,
		m_Coord_Y_B	,	HORIZONTAL_Y_MIN,	HORIZONTAL_Y_MAX	,
		m_Coord_X_C	,	VERTICAL_X_MIN	,	VERTICAL_X_MAX		,
		m_Coord_Y_C	,	VERTICAL_Y_MIN	,	VERTICAL_Y_MAX		,
		m_Coord_X_D	,	HORIZONTAL_X_MIN,	HORIZONTAL_X_MAX	,
		m_Coord_Y_D	,	HORIZONTAL_Y_MIN,	HORIZONTAL_Y_MAX	,
		m_Coord_X_E	,	HORIZONTAL_X_MIN,	HORIZONTAL_X_MAX	,
		m_Coord_Y_E	,	HORIZONTAL_Y_MIN,	HORIZONTAL_Y_MAX	,
		m_Coord_X_F	,	HORIZONTAL_X_MIN,	HORIZONTAL_X_MAX	,
		m_Coord_Y_F	,	HORIZONTAL_Y_MIN,	HORIZONTAL_Y_MAX	,
		m_Coord_X_G,	VERTICAL_X_MIN,		VERTICAL_X_MAX		,
		m_Coord_Y_G,	VERTICAL_Y_MIN,		VERTICAL_Y_MAX		,
		m_Coord_X_H,	HORIZONTAL_X_MIN,	HORIZONTAL_X_MAX	,
		m_Coord_Y_H,	HORIZONTAL_Y_MIN,	HORIZONTAL_Y_MAX	,
	};

	//個別表示データチェックテーブル
	std::string ConfigDisplayTbl[NM33_CONNECT_ENABLE_MAX] = {
		m_Display_A,
		m_Display_B,
		m_Display_C,
		m_Display_D,
		m_Display_E,
		m_Display_F,
		m_Display_G,
		m_Display_H,
	};

	sWord_Type sWGetConfig = 0;
	sWord_Type sWGetConfigDataMin = 0;
	sWord_Type sWGetConfigDataMax = 0;
	uWord_Type uWLoop = 0;
	uWord_Type uWLoopMax = (sizeof(ConfigDataChkTbl) / sizeof(ConfigData_t));

	//角度、座標調整用のコンフィギュレーションチェック
	for(uWLoop = 0; uWLoop < uWLoopMax; uWLoop++){
		sWGetConfig = ConfigDataChkTbl[uWLoop].sWConfig;
		sWGetConfigDataMin = ConfigDataChkTbl[uWLoop].sWConfigDataMin;
		sWGetConfigDataMax = ConfigDataChkTbl[uWLoop].sWConfigDataMax;

		if ((sWGetConfig < sWGetConfigDataMin) || (sWGetConfig > sWGetConfigDataMax)){
			//コンフィギュレーション設定エラー
			std::cout << MES_BIRDS_EYE_CONFIG_ERROR << std::endl;
			return RESULT_NG;
		}
	}

	//個別表示用のコンフィギュレーションチェック
	for(uWLoop = 0; uWLoop < NM33_CONNECT_ENABLE_MAX; uWLoop++){
		if((ConfigDisplayTbl[uWLoop] != DISPLAY_OFF) && (ConfigDisplayTbl[uWLoop] != DISPLAY_ON)){
			//コンフィギュレーション設定エラー
			std::cout << MES_BIRDS_EYE_CONFIG_ERROR << std::endl;
			return RESULT_NG;
		}
	}

	return RESULT_OK;
}

/*******************************************************************************
 *	Function Name		：BirdsEyeFPSLogOut
 *	Description			：俯瞰画像合成周期のFPSログ出力
 *	Argument			：なし
 *	Return				：なし
 *******************************************************************************/
void BirdsEyeComposer::BirdsEyeFPSLogOut(void)
{
	static coil::TimeValue LastTime = coil::gettimeofday();
	static uByte_Type uBCount = 0;

	if(uBCount > 100){
		uBCount = 0;
		coil::TimeValue Current_Time;
		Current_Time = coil::gettimeofday();
		double sec(Current_Time - LastTime);

		if (sec > 1.0 && sec < 1000.0){
			std::cout << "[MESSEGE]" << (100 / sec) << "FPS" << std::endl;
		}

		LastTime = Current_Time;
	}
	uBCount++;
}


extern "C"
{
 
  void BirdsEyeComposerInit(RTC::Manager* manager)
  {
    coil::Properties profile(birdseyecomposer_spec);
    manager->registerFactory(profile,
                             RTC::Create<BirdsEyeComposer>,
                             RTC::Delete<BirdsEyeComposer>);
  }
  
};


